@@ -18,6 +18,7 @@ import tempfile
 
 from oslo_concurrency import processutils as putils
 from oslo_log import log as logging
+from oslo_serialization import jsonutils
 from oslo_utils import fileutils
 from oslo_utils import netutils
 
@@ -71,11 +72,14 @@ class RBDConnector(base.BaseLinuxConnector):
 
     def _check_or_get_keyring_contents(self, keyring, cluster_name, user):
         try:
-            if keyring is None and user:
-                keyring_path = ("/etc/ceph/%s.client.%s.keyring" %
-                                (cluster_name, user))
-                with open(keyring_path, 'r') as keyring_file:
-                    keyring = keyring_file.read()
+            if keyring is None:
+                if user:
+                    keyring_path = ("/etc/ceph/%s.client.%s.keyring" %
+                                    (cluster_name, user))
+                    with open(keyring_path, 'r') as keyring_file:
+                        keyring = keyring_file.read()
+                else:
+                    keyring = ''
             return keyring
         except IOError:
             msg = (_("Keyring path %s is not readable.") % (keyring_path))
@@ -182,24 +186,95 @@ class RBDConnector(base.BaseLinuxConnector):
             # via the rbd kernel module.
             pool, volume = connection_properties['name'].split('/')
             rbd_dev_path = RBDConnector.get_rbd_device_name(pool, volume)
-            if (not os.path.islink(rbd_dev_path) or
-                    not os.path.exists(os.path.realpath(rbd_dev_path))):
+
+            if (
+                not os.path.islink(rbd_dev_path) or
+                not os.path.exists(os.path.realpath(rbd_dev_path))
+            ):
                 cmd = ['rbd', 'map', volume, '--pool', pool]
                 cmd += self._get_rbd_args(connection_properties)
                 self._execute(*cmd, root_helper=self._root_helper,
                               run_as_root=True)
             else:
-                LOG.debug('volume %(vol)s is already mapped to local'
-                          ' device %(dev)s',
-                          {'vol': volume,
-                           'dev': os.path.realpath(rbd_dev_path)})
+                LOG.debug(
+                    'Volume %(vol)s is already mapped to local device %(dev)s',
+                    {'vol': volume, 'dev': os.path.realpath(rbd_dev_path)}
+                )
 
-            return {'path': rbd_dev_path,
-                    'type': 'block'}
+            if (
+                not os.path.islink(rbd_dev_path) or
+                not os.path.exists(os.path.realpath(rbd_dev_path))
+            ):
+                LOG.warning(
+                    'Volume %(vol)s has not been mapped to local device '
+                    '%(dev)s; is the udev daemon running and are the '
+                    'ceph-renamer udev rules configured? See bug #1884114 for '
+                    'more information.',
+                    {'vol': volume, 'dev': rbd_dev_path},
+                )
+
+            return {'path': rbd_dev_path, 'type': 'block'}
 
         rbd_handle = self._get_rbd_handle(connection_properties)
         return {'path': rbd_handle}
 
+    def _find_root_device(self, connection_properties):
+        """Find the underlying /dev/rbd* device for a mapping.
+
+        Use the showmapped command to list all acive mappings and find the
+        underlying /dev/rbd* device that corresponds to our pool and volume.
+
+        :param connection_properties: The dictionary that describes all
+                                      of the target volume attributes.
+        :type connection_properties: dict
+        :returns: '/dev/rbd*' or None if no active mapping is found.
+        """
+        __, volume = connection_properties['name'].split('/')
+        cmd = ['rbd', 'showmapped', '--format=json']
+        cmd += self._get_rbd_args(connection_properties)
+        (out, err) = self._execute(*cmd, root_helper=self._root_helper,
+                                   run_as_root=True)
+
+        # ceph v13.2.0 (Mimic) changed the output format of 'rbd showmapped'
+        # from a dict of mappings keyed by ID to a simple list of mappings
+        # https://docs.ceph.com/docs/master/releases/mimic/
+        #
+        # before:
+        #
+        #   {
+        #     "0": {
+        #       "pool":"volumes",
+        #       "namespace":"",
+        #       "name":"volume-6d54cb90-a5d1-40d8-9cb2-c6adf43a02af",
+        #       "snap":"-",
+        #       "device":"/dev/rbd0"
+        #     }
+        #   }
+        #
+        # after:
+        #
+        #   [
+        #     {
+        #       "id":"0",
+        #       "pool":"volumes",
+        #       "namespace":"",
+        #       "name":"volume-6d54cb90-a5d1-40d8-9cb2-c6adf43a02af",
+        #       "snap":"-",
+        #       "device":"/dev/rbd0"
+        #     }
+        #   ]
+        #
+        # TODO(stephenfin): Drop when we drop support for ceph 13.2.0
+        mappings = jsonutils.loads(out)
+        if isinstance(mappings, dict):
+            # yes, we're losing the ID field but we don't need it here
+            mappings = mappings.values()
+
+        for mapping in mappings:
+            if mapping['name'] == volume:
+                return mapping['device']
+        return None
+
     @utils.trace
     def disconnect_volume(self, connection_properties, device_info,
                           force=False, ignore_errors=False):
@@ -214,12 +289,12 @@ class RBDConnector(base.BaseLinuxConnector):
         do_local_attach = connection_properties.get('do_local_attach',
                                                     self.do_local_attach)
         if do_local_attach:
-            pool, volume = connection_properties['name'].split('/')
-            dev_name = RBDConnector.get_rbd_device_name(pool, volume)
-            cmd = ['rbd', 'unmap', dev_name]
-            cmd += self._get_rbd_args(connection_properties)
-            self._execute(*cmd, root_helper=self._root_helper,
-                          run_as_root=True)
+            root_device = self._find_root_device(connection_properties)
+            if root_device:
+                cmd = ['rbd', 'unmap', root_device]
+                cmd += self._get_rbd_args(connection_properties)
+                self._execute(*cmd, root_helper=self._root_helper,
+                              run_as_root=True)
         else:
             if device_info:
                 rbd_handle = device_info.get('path', None)
