@@ -250,20 +250,70 @@ class RBDConnectorTestCase(test_connector.ConnectorTestCase):
 
         self.assertEqual(1, volume_close.call_count)
 
+    @ddt.data(
+        """
+        [{"id":"0","pool":"pool","device":"/dev/rbd0","name":"image"},
+         {"id":"1","pool":"pool","device":"/dev/rdb1","name":"image_2"}]
+        """,  # new-style output (ceph 13.2.0+)
+        """
+        {"0":{"pool":"pool","device":"/dev/rbd0","name":"image"},
+         "1":{"pool":"pool","device":"/dev/rdb1","name":"image_2"}}
+        """,  # old-style output
+    )
     @mock.patch.object(priv_rootwrap, 'execute', return_value=None)
-    def test_disconnect_local_volume(self, mock_execute):
+    def test_disconnect_local_volume(self, rbd_map_out, mock_execute):
+        """Test the disconnect volume case with local attach."""
         rbd_connector = rbd.RBDConnector(None, do_local_attach=True)
         conn = {'name': 'pool/image',
                 'auth_username': 'fake_user',
                 'hosts': ['192.168.10.2'],
                 'ports': ['6789']}
+        mock_execute.side_effect = [(rbd_map_out, None), (None, None)]
+        show_cmd = ['rbd', 'showmapped', '--format=json', '--id', 'fake_user',
+                    '--mon_host', '192.168.10.2:6789']
+        unmap_cmd = ['rbd', 'unmap', '/dev/rbd0', '--id', 'fake_user',
+                     '--mon_host', '192.168.10.2:6789']
+
         rbd_connector.disconnect_volume(conn, None)
 
-        dev_name = '/dev/rbd/pool/image'
-        cmd = ['rbd', 'unmap', dev_name, '--id', 'fake_user',
-               '--mon_host', '192.168.10.2:6789']
-        mock_execute.assert_called_once_with(*cmd, root_helper=None,
-                                             run_as_root=True)
+        # Assert that showmapped is used before we unmap the root device
+        mock_execute.assert_has_calls([
+            mock.call(*show_cmd, root_helper=None, run_as_root=True),
+            mock.call(*unmap_cmd, root_helper=None, run_as_root=True)])
+
+    @mock.patch.object(priv_rootwrap, 'execute', return_value=None)
+    def test_disconnect_local_volume_no_mapping(self, mock_execute):
+        rbd_connector = rbd.RBDConnector(None, do_local_attach=True)
+        conn = {'name': 'pool/not_mapped',
+                'auth_username': 'fake_user',
+                'hosts': ['192.168.10.2'],
+                'ports': ['6789']}
+        mock_execute.return_value = ("""
+{"0":{"pool":"pool","device":"/dev/rbd0","name":"pool-image"},
+ "1":{"pool":"pool","device":"/dev/rdb1","name":"pool-image_2"}}""", None)
+        show_cmd = ['rbd', 'showmapped', '--format=json', '--id', 'fake_user',
+                    '--mon_host', '192.168.10.2:6789']
+        rbd_connector.disconnect_volume(conn, None)
+
+        # Assert that only showmapped is called when no mappings are found
+        mock_execute.called_once_with(*show_cmd, root_helper=None,
+                                      run_as_root=True)
+
+    @mock.patch.object(priv_rootwrap, 'execute', return_value=None)
+    def test_disconnect_local_volume_no_mappings(self, mock_execute):
+        rbd_connector = rbd.RBDConnector(None, do_local_attach=True)
+        conn = {'name': 'pool/image',
+                'auth_username': 'fake_user',
+                'hosts': ['192.168.10.2'],
+                'ports': ['6789']}
+        mock_execute.return_value = ("{}", None)
+        show_cmd = ['rbd', 'showmapped', '--format=json', '--id', 'fake_user',
+                    '--mon_host', '192.168.10.2:6789']
+        rbd_connector.disconnect_volume(conn, None)
+
+        # Assert that only showmapped is called when no mappings are found
+        mock_execute.called_once_with(*show_cmd, root_helper=None,
+                                      run_as_root=True)
 
     def test_extend_volume(self):
         rbd_connector = rbd.RBDConnector(None)
